==> emelfm2-0.1.2.tar.gz: emelfm2-0.1.2/src/command/e2_alias.c <==
/* $Id: e2_alias.c 972 2005-07-29 00:49:46Z tpgww $
 * 
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelfm2.
 * emelfm2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelfm2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/command/e2_alias.c 
 * @brief functions for handling command aliases
 *
 * 'match' strings are compiled as extended regex.
 * If they do not start with '^', then one is prepended.
 * The whole alias list is scanned in order, so that aliases can be
 * chained, unless and until a match with its 'stop' flag set
 * 'replace' strings may contain:
 * \1 = substitute the matched alias
 * \2 = substitute that part of the input command string which is AFTER the matched alias
 * In theory,  \0 = substitute that part of the parsed string which is BEFORE
 * the matched alias, but actually the alias matching always starts from the start
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_alias.h"

/**
 * @brief cleanup data struct for @a alias
 * @param alias pointer to compiled-alias data struct
 * @return
 **/
static void _e2_alias_free (E2_Alias *alias)
{
	g_free (alias->match);
	g_free (alias->replace);
	regfree (&(alias->preg));
	g_free (alias);
}
/**
 * @brief create data struct for an alias entry
 * Back-reference flag is set if @a replace includes \1 or \2
 * \0 is now irrelevant
 * @param match string to be replaced
 * @param replace string to use as a replacement
 * @param preg pointer to regex struct
 * @param stop TRUE if the scan is to stop at the first match
 * @return pointer to alias struct with the matching data
 **/
static E2_Alias *_e2_alias_get (gchar *match, gchar *replace, regex_t preg,
	gboolean stop)
{
	E2_Alias *alias = g_malloc (sizeof (E2_Alias));
	alias->match = match;
	alias->replace = replace;
	alias->preg = preg;
	alias->stop = stop;
	
	//test if the replace string has a back reference
	alias->has_back_reference = FALSE;
	gchar *test = NULL;
	if ((test = strchr (replace, '\\')) != NULL)	//if always ascii \, don't need g_utf8_strchr()
	{
		if (
		//	test[1] == '0' || 
		test[1] == '1' || test[1] == '2')
			alias->has_back_reference = TRUE;
	}
	return alias;
}
/**
 * @brief create replacement for @a str with alias replaced
 * This tries to match from start of string, as compiled aliases start
 * with ^ and were not compiled with flag REG_NOTBOL
 * @param alias pointer to compiled-alias data struct
 * @param str command string, the start of which might match an alias
 * @param replaced location for storing the string with the alias substituted
 * @return TRUE if a match was found, and if so, a newly allocated string in @a replaced
 **/
static gboolean _e2_alias_match (E2_Alias *alias, gchar *str, gchar **replaced)
{
	printd (DEBUG, "match alias for %s", str);
	regmatch_t pmatch;
	gint ret = regexec (&alias->preg, str, 1, &pmatch, 0);
	gchar *tmp = NULL, *tmp2 = NULL;
	if (ret == REG_NOMATCH)
		return FALSE;

	else
		if (ret == 0)
	{
		if ((pmatch.rm_eo == -1) || (pmatch.rm_so == -1))
		//the compiled string wasn't acually used ??
			return FALSE;
		//make sure the whole of the command matches (ie avoid e.g. x matches xpdf)
//		if (*(str+pmatch.rm_eo) != ' ' && *(str+pmatch.rm_eo) != '\0')
//			return FALSE;
		//so we have a match, replace it
		if (alias->has_back_reference)
		{
			tmp = g_strdup (alias->replace);
//			printd (DEBUG, "match alias ended at offset %d", pmatch.rm_eo);
			//carve off the end of the source string
			if (pmatch.rm_eo > 0)
			{
				gchar *real_start = str + pmatch.rm_eo;
//				if (*real_start != '\0')
//					real_start--;	//backup so we grab the (last) separator too
				tmp2 = e2_utils_str_replace (tmp, "\\2", real_start);
			}
			else
				tmp2 = e2_utils_str_replace (tmp, "\\2", "");
			
			g_free (tmp);
			tmp = tmp2;

/*			//carve off any prior part of the source string
			//(can never happen while ^ at start)
			if (pmatch.rm_so > 0)
			{
				gchar *match = g_strndup (str, pmatch.rm_so);
				tmp2 = e2_utils_str_replace (tmp, "\\0", match);
				g_free (match);
			}
			else
				tmp2 = e2_utils_str_replace (tmp, "\\0", "");

			g_free (tmp);
			tmp = tmp2;
*/
			//build the replacement string with the alias in place
			gchar *match = g_strndup (str + pmatch.rm_so, pmatch.rm_eo - pmatch.rm_so);
			*replaced = e2_utils_str_replace (tmp, "\\1", match);

			g_free (tmp);
			g_free (match);
		}
		else	//no back-ref
		{
			gchar *last = str + pmatch.rm_eo;
//			if (*last != '\0')
//				last--;	//backup so we grab the (last) separator too

/*			if (pmatch.rm_so > 0)   //see above - never pass this test
			{
				gchar *first = g_strndup (str, pmatch.rm_so);
				*replaced = g_strconcat (first, alias->replace, last, NULL);
				g_free (first);
			} else
*/				*replaced = g_strconcat (alias->replace, last, NULL);
		}
		return TRUE;
	} else
	{
		printd (DEBUG, "regex error");
		return FALSE;
	}
}
/**
 * @brief convert all alias data into a form ready for use
 * This adds a '^' to the beginning of the alias if not there already 
 * (causing matching from the start) so that the config data doesn't
 * need to show a '^'. This means that aliases can only work for 
 * a command, not for a part of a command after its start.
 * To avoid matches like x <=> xpdf, a separator is added to 
 * the end of each alias.
 * The match strings are 'compiled' as extended regex.
 * Data structs are put into a list.
 * @param rt pointer to alias data struct
 * @param iter pointer to tree iter, used for option tree traversal
 * @return
 **/
static void _e2_alias_add_all (E2_AliasRuntime *rt, GtkTreeIter *iter)
{
	do
	{
		gchar *match, *replace, 	*builder;
		gboolean stop;
		gtk_tree_model_get (rt->set->ex.tree.model, iter,
			0, &match, 1, &stop, 2, &replace, -1);

		if ((match != NULL) && (*match != '\0'))
		{
			if (*match != '^')
			{
				builder = g_strconcat ("^", match, NULL);
				g_free (match);
				match = builder;
			}
			//add separator
			builder = g_strconcat (match, "([ \t]+|$)", NULL);
			g_free (match);
			match = builder;
			regex_t preg;
			if (!regcomp (&preg, match, REG_EXTENDED))
			{
				E2_Alias *alias = _e2_alias_get (match, replace, preg, stop);
				rt->aliases = g_list_append (rt->aliases, alias);
			}
			else
				printd (WARN, "error compiling regexp pattern '%s'", match);
		}
	} while (gtk_tree_model_iter_next (rt->set->ex.tree.model, iter));
}

  /******************/
 /***** public *****/
/******************/

/**
 * @brief find a matching alias for @a str
 * If the aliases have not been enlisted, then that is done first.
 * The whole alias list is scanned in order, so that aliases can be chained,
 * unless and until a match has its 'stop' flag set
 * @param str a string starting with a command for which there may be an alias
 * @param rt pointer to aliases runtime data struct
 * @return a newly-allocated string, either the matched alias, or the same as @a str if there is no match
 **/
gchar *e2_alias_apply (gchar *str, E2_AliasRuntime *rt)
{
	str = g_strdup (str);	//never touch the original string
	if (!rt->sync)
	{
		e2_alias_sync (rt);
		if (!rt->sync)
			return str;
	}
//	gint cmdlen = strlen (str);
	gchar *ret = NULL;  //no-match signal

	GList *list = g_list_first (rt->aliases);
	while (list != NULL)
	{
		if (_e2_alias_match (list->data, str, &ret))   //cmdlen, &ret))
		{
			g_free (str);
			str = ret;
			E2_Alias *alias = list->data;
			if (alias->stop)
				break;
		}
		list = g_list_next (list);
	}
	
	if (ret == NULL)
		return g_strdup (str);
	else
		return ret;
}
/**
 * @brief refresh alias data
 * Any aliases in the list are cleared, a replacement list is built
 * from config data if possible, and the 'sync' flag is set accordingly
 * @param rt pointer to aliases runtime data struct
 * @return
 **/
void e2_alias_sync (E2_AliasRuntime *rt)
{
	printd (DEBUG, "e2_alias_sync (rt:)");
	if (rt->aliases != NULL)
	{
		g_list_foreach (rt->aliases, (GFunc) _e2_alias_free, NULL);
		g_list_free (rt->aliases);
		rt->aliases = NULL;
	}
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (rt->set->ex.tree.model, &iter))
	{
		_e2_alias_add_all (rt, &iter);
		rt->sync = TRUE;
	}
	else
	{
		printd (NOTICE, "no aliases defined");
		rt->sync = FALSE;
	}
}
/**
 * @brief initialize alias-related options, and init some alias flags
 * @return
 **/
void e2_alias_init (void)
{
	E2_AliasRuntime *rt = &app.aliases;
	rt->aliases = NULL;  //moved up from end
	rt->sync = FALSE;
	//this option does not require a rebuild after config change
	gchar *group_name = g_strconcat(_C(8),".",_C(0),NULL); //_("commands.aliases"
	printd (DEBUG, "e2_alias_init (rt:,group:%s)", group_name);
	rt->set = e2_option_tree_register ("command-aliases", group_name, _C(0),
		NULL, NULL, NULL, E2_OPTION_TREE_UP_DOWN | E2_OPTION_TREE_ADD_DEL |
		E2_OPTION_TREE_LIST,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP | E2_OPTION_FLAG_BUILDALIAS);
	e2_option_tree_add_column (rt->set, _("Match"), E2_OPTION_TREE_TYPE_STR,
		0, "", NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (rt->set, _("Stop"), E2_OPTION_TREE_TYPE_BOOL,
		FALSE, "false", NULL, NULL, 0, NULL, NULL); //this stops alias scan when 1st match found
	e2_option_tree_add_column (rt->set, _("Replace"), E2_OPTION_TREE_TYPE_STR,
		0, "", NULL, NULL, 0, NULL, NULL);
	e2_option_tree_sync_columns (rt->set);
	
	_e2_option_tree_prepare_default (rt->set,
		g_strdup("<command-aliases"),  //internal name
		g_strdup("[^%].*[<>\\|].*||>\\1"),	//make sure re-directions and pipes are run in a new shell
//		g_strdup(";\\|\\\\|\\|>\\|<\\|\\*\\|=||%\\0\\1\\2"),
		//some terminals (eg gnome-terminal) don't support these
		//NB $TERM doesn't always exist, or might be invalid
//		g_strconcat(_("x"),"|true|$[command-xterm] -e sh -c '\\2'",NULL),
//		g_strconcat(_("xx"),"|true|$[command-xterm] -e sh -c '\\2;echo -n \"",_("Done. Press enter "),"\";read'",NULL),
//		g_strconcat(_("su"),"|true|$[command-xterm] -e sh -c 'su -c \\2;echo -n \"",_("Done. Press enter "),"\";read'",NULL),
		g_strconcat(_("x"),"|true|xterm -e sh -c '\\2'",NULL),
		g_strconcat(_("xx"),"|true|xterm -e sh -c '\\2;echo -n \"",_("Done. Press enter "),"\";read'",NULL),
		g_strconcat(_("su"),"|true|xterm -e sh -c 'su -c \\2;echo -n \"",_("Done. Press enter "),"\";read'",NULL),
		g_strconcat(_("clear"),"|true|<",_A(8),".",_A(29),NULL),
		g_strconcat(_("quit"),"|true|<",_A(1),".",_A(83),NULL),	//command.quit
		g_strconcat(_("help"),"|true|<",_A(5),".",_A(76)," "DOC_DIR, MAIN_HELP" [",_("commands"),"]",NULL),  //file.view_at
		g_strconcat(_("keys"),"|true|<",_A(8),".",_A(45)," ",_("keys")," \\2",NULL),
		g_strconcat(_("e2ps"),"|true|<",_A(1),".",_A(47),NULL),
		g_strconcat(_("wget"),"|true|\\1 --progress=bar:force \\2",NULL),
		g_strdup(">"),
		NULL);
}

==> emelfm2-0.1.4.tar.gz: emelfm2-0.1.4/src/command/e2_alias.c <==
/* $Id: e2_alias.c 1433 2005-12-13 00:36:23Z tpgww $
 * 
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelfm2.
 * emelfm2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelfm2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/command/e2_alias.c 
 * @brief functions for handling command aliases
 *
 * 'match' strings are compiled as extended regex.
 * If they do not start with '^', then one is prepended.
 * The whole alias list is scanned in order, so that aliases can be
 * chained, unless and until a match with its 'stop' flag set
 * 'replace' strings may contain:
 * \1 = substitute the matched alias
 * \2 = substitute that part of the input command string which is AFTER the matched alias
 * In theory,  \0 = substitute that part of the parsed string which is BEFORE
 * the matched alias, but actually the alias matching always starts from the start
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_alias.h"

/**
 * @brief cleanup data struct for @a alias
 * @param alias pointer to compiled-alias data struct
 * @return
 **/
static void _e2_alias_free (E2_Alias *alias)
{
	g_free (alias->match);
	g_free (alias->replace);
	regfree (&(alias->preg));
	g_free (alias);
}
/**
 * @brief create data struct for an alias entry
 * Back-reference flag is set if @a replace includes \1 or \2
 * \0 is now irrelevant
 * @param match string to be replaced
 * @param replace string to use as a replacement
 * @param preg pointer to regex struct
 * @param stop TRUE if the scan is to stop at the first match
 * @return pointer to alias struct with the matching data
 **/
static E2_Alias *_e2_alias_get (gchar *match, gchar *replace, regex_t preg,
	gboolean stop)
{
	E2_Alias *alias = g_malloc (sizeof (E2_Alias));
	alias->match = match;
	alias->replace = replace;
	alias->preg = preg;
	alias->stop = stop;
	
	//test if the replace string has a back reference
	alias->has_back_reference = FALSE;
	gchar *test = NULL;
	if ((test = strchr (replace, '\\')) != NULL)	//if always ascii \, don't need g_utf8_strchr()
	{
		if (
		//	test[1] == '0' || 
		test[1] == '1' || test[1] == '2')
			alias->has_back_reference = TRUE;
	}
	return alias;
}
/**
 * @brief create replacement for @a str with alias replaced
 * This tries to match from start of string, as compiled aliases start
 * with ^ and were not compiled with flag REG_NOTBOL
 * @param alias pointer to compiled-alias data struct
 * @param str command string, the start of which might match an alias
 * @param replaced location for storing the string with the alias substituted
 * @return TRUE if a match was found, and if so, a newly allocated string in @a replaced
 **/
static gboolean _e2_alias_match (E2_Alias *alias, gchar *str, gchar **replaced)
{
	printd (DEBUG, "match alias for %s", str);
	regmatch_t pmatch;
	gint ret = regexec (&alias->preg, str, 1, &pmatch, 0);
	gchar *tmp = NULL, *tmp2 = NULL;
	if (ret == REG_NOMATCH)
		return FALSE;

	else
		if (ret == 0)
	{
		if ((pmatch.rm_eo == -1) || (pmatch.rm_so == -1))
		//the compiled string wasn't acually used ??
			return FALSE;
		//make sure the whole of the command matches (ie avoid e.g. x matches xpdf)
//		if (*(str+pmatch.rm_eo) != ' ' && *(str+pmatch.rm_eo) != '\0')
//			return FALSE;
		//so we have a match, replace it
		if (alias->has_back_reference)
		{
			tmp = g_strdup (alias->replace);
//			printd (DEBUG, "match alias ended at offset %d", pmatch.rm_eo);
			//carve off the end of the source string
			if (pmatch.rm_eo > 0)
			{
				gchar *real_start = str + pmatch.rm_eo;
//				if (*real_start != '\0')
//					real_start--;	//backup so we grab the (last) separator too
				tmp2 = e2_utils_str_replace (tmp, "\\2", real_start);
			}
			else
				tmp2 = e2_utils_str_replace (tmp, "\\2", "");
			
			g_free (tmp);
			tmp = tmp2;

/*			//carve off any prior part of the source string
			//(can never happen while ^ at start)
			if (pmatch.rm_so > 0)
			{
				gchar *match = g_strndup (str, pmatch.rm_so);
				tmp2 = e2_utils_str_replace (tmp, "\\0", match);
				g_free (match);
			}
			else
				tmp2 = e2_utils_str_replace (tmp, "\\0", "");

			g_free (tmp);
			tmp = tmp2;
*/
			//build the replacement string with the alias in place
			gchar *match = g_strndup (str + pmatch.rm_so, pmatch.rm_eo - pmatch.rm_so);
			*replaced = e2_utils_str_replace (tmp, "\\1", match);

			g_free (tmp);
			g_free (match);
		}
		else	//no back-ref
		{
			gchar *last = str + pmatch.rm_eo;
//			if (*last != '\0')
//				last--;	//backup so we grab the (last) separator too

/*			if (pmatch.rm_so > 0)   //see above - never pass this test
			{
				gchar *first = g_strndup (str, pmatch.rm_so);
				*replaced = g_strconcat (first, alias->replace, last, NULL);
				g_free (first);
			} else
*/				*replaced = g_strconcat (alias->replace, last, NULL);
		}
		return TRUE;
	} else
	{
		printd (DEBUG, "regex error");
		return FALSE;
	}
}
/**
 * @brief convert all alias data into a form ready for use
 * This adds a '^' to the beginning of the alias if not there already 
 * (causing matching from the start) so that the config data doesn't
 * need to show a '^'. This means that aliases can only work for 
 * a command, not for a part of a command after its start.
 * To avoid matches like x <=> xpdf, a separator is added to 
 * the end of each alias.
 * The match strings are 'compiled' as extended regex.
 * Data structs are put into a list.
 * @param rt pointer to alias data struct
 * @param iter pointer to tree iter, used for option tree traversal
 * @return
 **/
static void _e2_alias_add_all (E2_AliasRuntime *rt, GtkTreeIter *iter)
{
	do
	{
		gchar *match, *replace, 	*builder;
		gboolean stop;
		gtk_tree_model_get (rt->set->ex.tree.model, iter,
			0, &match, 1, &stop, 2, &replace, -1);

		if ((match != NULL) && (*match != '\0'))
		{
			if (*match != '^')
			{
				builder = g_strconcat ("^", match, NULL);
				g_free (match);
				match = builder;
			}
			//add separator
			builder = g_strconcat (match, "([ \t]+|$)", NULL);
			g_free (match);
			match = builder;
			regex_t preg;
			if (!regcomp (&preg, match, REG_EXTENDED))
			{
				E2_Alias *alias = _e2_alias_get (match, replace, preg, stop);
				rt->aliases = g_list_append (rt->aliases, alias);
			}
			else
				printd (WARN, "error compiling regexp pattern '%s'", match);
		}
	} while (gtk_tree_model_iter_next (rt->set->ex.tree.model, iter));
}

  /******************/
 /***** public *****/
/******************/

/**
 * @brief find a matching alias for @a str
 * If the aliases have not been enlisted, then that is done first.
 * The whole alias list is scanned in order, so that aliases can be chained,
 * unless and until a match has its 'stop' flag set
 * @param str a string starting with a command for which there may be an alias
 * @param rt pointer to aliases runtime data struct
 * @return a newly-allocated string, either the matched alias, or the same as @a str if there is no match
 **/
gchar *e2_alias_apply (gchar *str, E2_AliasRuntime *rt)
{
	str = g_strdup (str);	//never touch the original string
	if (!rt->sync)
	{
		e2_alias_sync (rt);
		if (!rt->sync)
			return str;
	}
//	gint cmdlen = strlen (str);
	gchar *ret = NULL;  //no-match signal

	GList *list = g_list_first (rt->aliases);
	while (list != NULL)
	{
		if (_e2_alias_match (list->data, str, &ret))   //cmdlen, &ret))
		{
			g_free (str);
			str = ret;
			E2_Alias *alias = list->data;
			if (alias->stop)
				break;
		}
		list = g_list_next (list);
	}
	
	if (ret == NULL)
		return g_strdup (str);
	else
		return ret;
}
/**
 * @brief refresh alias data
 * Any aliases in the list are cleared, a replacement list is built
 * from config data if possible, and the 'sync' flag is set accordingly
 * @param rt pointer to aliases runtime data struct
 * @return
 **/
void e2_alias_sync (E2_AliasRuntime *rt)
{
	printd (DEBUG, "e2_alias_sync (rt:)");
	if (rt->aliases != NULL)
	{
		g_list_foreach (rt->aliases, (GFunc) _e2_alias_free, NULL);
		g_list_free (rt->aliases);
		rt->aliases = NULL;
	}
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (rt->set->ex.tree.model, &iter))
	{
		_e2_alias_add_all (rt, &iter);
		rt->sync = TRUE;
	}
	else
	{
		printd (NOTICE, "no aliases defined");
		rt->sync = FALSE;
	}
}
/**
 * @brief initialize alias-related options, and init some alias flags
 * @return
 **/
void e2_alias_init (void)
{
	E2_AliasRuntime *rt = &app.aliases;
	rt->aliases = NULL;  //moved up from end
	rt->sync = FALSE;
	//this option does not require a rebuild after config change
	gchar *group_name = g_strconcat(_C(5),".",_C(0),NULL); //_("commands.aliases"
	printd (DEBUG, "e2_alias_init (rt:,group:%s)", group_name);
	rt->set = e2_option_tree_register ("command-aliases", group_name, _C(0),
		NULL, NULL, NULL, E2_OPTION_TREE_UP_DOWN | E2_OPTION_TREE_ADD_DEL |
		E2_OPTION_TREE_LIST,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP | E2_OPTION_FLAG_BUILDALIAS);
	e2_option_tree_add_column (rt->set, _("Match"), E2_OPTION_TREE_TYPE_STR,
		0, "", NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (rt->set, _("Stop"), E2_OPTION_TREE_TYPE_BOOL,
		FALSE, "false", NULL, NULL, 0, NULL, NULL); //this stops alias scan when 1st match found
	e2_option_tree_add_column (rt->set, _("Replace"), E2_OPTION_TREE_TYPE_STR,
		0, "", NULL, NULL, 0, NULL, NULL);
	e2_option_tree_sync_columns (rt->set);
	
	_e2_option_tree_prepare_default (rt->set,
		g_strdup("<command-aliases"),  //internal name
		g_strdup("[^%].*[<>\\|].*||>\\1"),	//make sure re-directions and pipes are run in a new shell
//		g_strdup(";\\|\\\\|\\|>\\|<\\|\\*\\|=||%\\0\\1\\2"),
		//some terminals (eg gnome-terminal) don't support these
		//NB $TERM doesn't always exist, or might be invalid
//		g_strconcat(_("x"),"|true|$[command-xterm] -e sh -c '\\2'",NULL),
//		g_strconcat(_("xx"),"|true|$[command-xterm] -e sh -c '\\2;echo -n \"",_("Done. Press enter "),"\";read'",NULL),
//		g_strconcat(_("su"),"|true|$[command-xterm] -e sh -c 'su -c \\2;echo -n \"",_("Done. Press enter "),"\";read'",NULL),
		g_strconcat(_("x"),"|true|xterm -e sh -c '\\2'",NULL),
		g_strconcat(_("xx"),"|true|xterm -e sh -c '\\2;echo -n \"",_("Done. Press enter "),"\";read'",NULL),
		g_strconcat(_("su"),"|true|xterm -e sh -c 'su -c \"\\2\";echo -n \"",_("Done. Press enter "),"\";read'",NULL),
		g_strconcat(_("clear"),"|true|<",_A(8),".",_A(26),NULL),
		g_strconcat(_("quit"),"|true|<",_A(1),".",_A(60),NULL),	//command.quit
		g_strconcat(_("help"),"|true|<",_A(5),".",_A(82)," "DOC_DIR, MAIN_HELP" [",_("commands"),"]",NULL),  //file.view_at
		g_strconcat(_("keys"),"|true|<",_A(8),".",_A(43)," ",_("keys")," \\2",NULL),
		g_strconcat(_("e2ps"),"|true|<",_A(1),".",_A(46),NULL),
		g_strconcat(_("wget"),"|true|\\1 --progress=bar:force \\2",NULL),
		g_strdup(">"),
		NULL);
}
